import hashlib
import re
from functools import reduce
from typing import List, Any, Dict, Union

import numpy as np
from pandas import DataFrame

from sktmls.meta_tables import MetaTable
from sktmls.models import MLSCatBoostModel, MLSModelError


class CommbotCatboostDeviceModel(MLSCatBoostModel):
    """
    Commbot CatBoost 기반 클래스입니다.
    주어진 `features`리스트를 이용해 prediction 후 스코어 상위 3개 단말그룹으로 필터링 합니다.
    이후 `id`, `name`, `score`, `props`, `type`을 반환하는 모델입니다.
    """

    def __init__(
        self,
        model,
        model_name: str,
        model_version: str,
        model_features: List[str],
        default_model_feature_values: List[Any],
        target_prod_id: List[str],
        popular_devices: List[str],
        products_meta: Dict[str, Any],
        products: Dict[str, Any],
        device_meta: MetaTable,
        context_meta: DataFrame,
        conversion_formulas: Dict[str, Dict[str, Any]] = {},
        emb_features: List[str] = [],
        default_emb_feature_values: List[List[Any]] = [],
        emb_feature_indices: List[List] = [],
        default_context_value: str = "context_default",
        num_pick: int = 3,
    ):
        """
        ## Args
        - model: Catboost로 학습한 모델 객체
        - model_name: (str) 모델 이름
        - model_version: (str) 모델 버전
        - model_features: (list(str)) non-embedding 피쳐 리스트(필수 피쳐: `eqp_mdl_cd`, `age`)
        - default_model_feature_values: (list) 피쳐 기본값 리스트
        - target_prod_id: (list) 타겟 상품 ID 리스트
        - popular_devices: (list) 인기 디바이스 리스트
        - products: (dict) 컨텍스트 값
        - products_meta: (dict) 상품 메타 정보
        - device_meta: (`sktmls.meta_tables.MetaTable`) 단말코드그룹 값
        - context_meta: (DataFrame) 컨텍스트 메타 정보
        - conversion_formulas: (optional) (dict(dict)) Feature conversion에 사용할 조건 dict. 형식은 아래 Example 참조 (기본값: {})
        - emb_features: (optional) (list(str)) 임베딩 피쳐 이름 리스트 (기본값: [])
        - default_emb_feature_values: (optional) (list(list)) 임베딩 피쳐들의 기본값 리스트 (리스트의 리스트 형식) (기본값: [])
        - emb_feature_indices: (optional) (list(list)) 각 임베딩 피쳐의 세부 사용 피쳐 인덱스 리스트 (기본값: [])
        - default_context_value: (optional) (str) 기본 컨텍스트 값 (기본값: "context_default")
        ## Example
        ```python
        model_features = ["feature1", "feature2", "feature3"]
        default_model_feature_values = [
            dimension_client.get_dimension(dimension_type="user", name=feature).default for feature in model_features
        ]
        device_meta = metatable_client.get_meta_table(name="test_meta_table")
        context_meta = metatable_client.get_meta_table_dataframe(name="test_context_meta")
        products_meta={
            "test_prod_id": {"product_grp_nm": "test_product_grp_nm", "mfact_nm": "test_mfact_nm", "eqp_mdl_ntwk": "4G"},
        }
        popular_devices = ["popular001", "popular002", "popular003" ]
        products = {
            "test_prod_id": {
                "name": "test_prod_name",
                "type": "device",
                "context_features": ["context_default"],
                "random_context": False,
            }
        }
        my_model_v1 = CommbotCatboostDeviceModel(
            model=catboost_model,
            model_name="my_model",
            model_version="v1",
            model_features=["user_id", "eqp_mdl_cd", "age"] + model_features,
            default_model_feature_values=default_model_feature_values,
            products_meta=products_meta,
            products = products,
            target_prod_id = ["prod_1", "prod_2"],
            popular_devices=popular_devices,
            device_meta=device_meta,
            conversion_formulas=conversion_formulas,
            context_meta=context_meta,
            emb_features=["embedding_vector"],
            default_emb_feature_values=[[0.0] * 64],
            emb_feature_indices=[[0, 1]],
        )
        result = my_model_v1.predict(["eqp_mdl_cd", "value_1", "value_2", "value_3"], keys=[[], []])
        """

        assert isinstance(model_features, list), "`model_features`는 list타입이어야 합니다."
        assert (
            model_features[0] == "user_id" and model_features[1] == "eqp_mdl_cd" and model_features[2] == "age"
        ), "`user_id`, `eqp_mdl_cd`, `age`는 필수 피쳐 입니다."
        assert isinstance(default_model_feature_values, list), "`default_model_feature_values`는 list타입이어야 합니다."
        assert len(model_features) - 3 == len(
            default_model_feature_values
        ), "`model_features`(필수 피쳐제외)와 `default_model_feature_values`의 길이가 다릅니다."
        assert None not in default_model_feature_values, "`default_model_feature_values`에 None이 포함되어 있습니다."

        assert isinstance(emb_features, list), "`emb_features`는 리스트 형식이어야 합니다."
        assert isinstance(default_emb_feature_values, list), "`default_emb_feature_values`는 리스트 형식이어야 합니다."
        assert None not in default_emb_feature_values, "`default_emb_feature_values`에 None이 포함되어 있습니다."
        assert isinstance(emb_feature_indices, list), "`emb_feature_indices`는 리스트 형식이어야 합니다."
        assert None not in emb_feature_indices, "`emb_feature_indices`에 None이 포함되어 있습니다."
        assert (
            len(emb_features) == len(default_emb_feature_values) == len(emb_feature_indices)
        ), "`emb_features`, `default_emb_feature_values`, `emb_feature_indices`의 길이가 다릅니다."

        for i, indices in enumerate(emb_feature_indices):
            assert isinstance(indices, list) and isinstance(
                default_emb_feature_values[i], list
            ), "`emb_feature_indices`와 `default_emb_feature_values` 내 모든 element는 리스트 형식이어야 합니다."
            assert len(indices) <= len(
                default_emb_feature_values[i]
            ), "`emb_feature_indices` 내 element 길이가 해당 임베딩 피쳐 길이보다 깁니다."
            assert None not in default_emb_feature_values[i], "`default_emb_feature_values` 내 element에 None이 포함되어 있습니다."
            for idx in indices:
                assert isinstance(idx, int), "`emb_feature_indices` 내 모든 인덱스 값은 int 형식이어야 합니다."
                assert idx < len(default_emb_feature_values[i]), "`emb_feature_indices` 내 인덱스가 임베딩 피쳐 전체 길이보다 깁니다."

        assert isinstance(target_prod_id, list) and len(target_prod_id) > 0, "`target_prod_id`는 리스트 형식이어야 하며 필수 입력값입니다."
        assert isinstance(products_meta, dict) and len(products_meta) > 0, "`products_meta`는 딕셔너리 형식이어야 하며 필수 입력값입니다."
        assert isinstance(products, dict) and len(products) > 0, "`products`는 딕셔너리 형식이어야 하며 필수 입력값입니다."
        assert isinstance(device_meta, MetaTable) and len(device_meta.items) > 0, "`device_meta`는 필수 입력 값 입니다."
        assert isinstance(context_meta, DataFrame) and len(context_meta) > 0, "`context_meta`는 필수 입력 값 입니다."
        for column in ["ContextId", "ItemNm", "Text ID", "T world text"]:
            assert column in context_meta.columns, "`context_meta`의 필수 컬럼이 누락되어 있습니다"
        assert isinstance(popular_devices, list) and len(popular_devices) > 0, "`popular_devices`는 필수 입력값입니다."

        assert isinstance(conversion_formulas, dict), "`conversion_formulas`는 dict 형식이어야 합니다."
        for k, v in conversion_formulas.items():
            assert isinstance(k, str), "`conversion_formulas`의 키는 문자열이어야 합니다."
            assert isinstance(v, dict), "`conversion_formulas`의 값은 dict 형식이어야 합니다."
            assert "map" in v or "conditions" in v, "`conversion_formulas`의 값에 map 또는 conditions 키가 포함되어야 합니다."
            assert "default" in v, "`conversion_formulas`의 값에 default 키가 포함되어야 합니다."

        for device_meta_item in device_meta.items:
            device_meta_item_name = device_meta_item.get("name")
            device_meta_item_value = device_meta_item.get("values")

            assert (
                device_meta_item_value and device_meta_item_name
            ), "`device_meta`의 `item` 항목 중 `name`과 `values`는 필수입니다."
            assert isinstance(device_meta_item_name, str), "`device_meta.item`의 `name`은 str 타입이어야 합니다."
            assert isinstance(device_meta_item_value, dict), "`device_meta.item`의 `values`은 dict 타입이어야 합니다."
            assert {
                "product_grp_id",
                "product_grp_nm",
                "mfact_nm",
            } <= device_meta_item_value.keys(), (
                "`device_meta.item`의 `values`는 `product_grp_id`, `product_grp_nm`, `mfact_nm` 키를 포함하여야 합니다."
            )

        super().__init__(model, model_name, model_version, model_features + emb_features)
        self.default_model_feature_values = default_model_feature_values
        self.target_prod_id = target_prod_id
        self.device_meta = {meta["name"]: meta["values"] for meta in device_meta.items}
        self.context_meta = context_meta[context_meta["ItemNm"] == "device nm"]
        self.conversion_formulas = conversion_formulas
        self.products_meta = products_meta
        self.products = products
        self.default_emb_feature_values = default_emb_feature_values
        self.emb_feature_indices = emb_feature_indices
        self.default_context_value = default_context_value
        self.popular_devices = popular_devices
        self.num_pick = num_pick

    def predict(
        self,
        x: List[Any],
        user_id: str = hashlib.sha256("default".encode("utf-8")).hexdigest(),
        **kwargs,
    ) -> Dict[str, Any]:
        # 1. 전처리
        model_x = self._preprocess(x[3:])

        # 2. ML Prediction
        y = self._ml_predict(model_x)

        # 3. 후처리
        items = self._postprocess(y, x[0], x[1], x[2], keys=kwargs.get("keys", [[], []])) or []
        return {"items": items}

    def _preprocess(self, x: List[Any]) -> List[Any]:
        # 1. 전처리
        try:
            if len(self.features) - 3 != len(x):
                raise MLSModelError("CommbotCatboostDeviceModel: `x`와 `features`의 길이가 다릅니다.")

            # - 1.1 피쳐 나누기
            emb_x = x[len(x) - len(self.default_emb_feature_values) :]
            for i, emb_f in enumerate(emb_x):
                if not emb_f:
                    emb_x[i] = self.default_emb_feature_values[i]
                    continue
                if not isinstance(emb_f, list):
                    raise MLSModelError("CommbotCatboostDeviceModel: emb_x는 list 타입이어야 합니다.")
                if len(emb_f) != len(self.default_emb_feature_values[i]):
                    raise MLSModelError(
                        "CommbotCatboostDeviceModel: emb_x와 `default_emb_feature_values`의 임베딩 피쳐 길이가 다릅니다."
                    )
                if len(emb_f) < len(self.emb_feature_indices[i]):
                    raise MLSModelError(
                        "CommbotCatboostDeviceModel: `emb_feature_indices`의 임베딩 피쳐 인덱스 길이가 emb_x 임베딩 피쳐 길이보다 큽니다."
                    )

            # - 1.2. None 피쳐 기본값 할당 및 임베딩 피쳐 펼치기
            non_emb_x = x[: len(x) - len(self.default_emb_feature_values)]
            non_emb_x = [f if f is not None else self.default_model_feature_values[i] for i, f in enumerate(non_emb_x)]
            emb_x = [
                [
                    emb_x[i][j] if emb_x[i][j] is not None else self.default_emb_feature_values[i][j]
                    for j in self.emb_feature_indices[i]
                ]
                for i in range(len(emb_x))
            ]
            emb_x = reduce(lambda a, b: a + b, emb_x, [])

            return self._convert(non_emb_x) + emb_x
        except Exception as e:
            raise MLSModelError(f"CommbotCatboostDeviceModel: 전처리에 실패했습니다. {e}")

    def _ml_predict(self, x: List[Any]) -> List[float]:
        try:
            y = self.models[0].predict_proba(np.array([x]))
            if len(y) == 1:
                y = y[0]
            try:
                return y.tolist()
            except AttributeError:
                return y
        except Exception as e:
            raise MLSModelError(f"CommbotCatboostDeviceModel: ML Prediction에 실패했습니다. {e}")

    def _postprocess(
        self, y: Union[List[float], None], user_id: str, user_eqp_mdl_cd: str, user_age: int, keys: list
    ) -> List[Dict[str, Any]]:
        try:
            num_pick = min(self.num_pick, len(y))
            prefer_mfact_list = keys[1]

            if user_id == hashlib.sha256("default".encode("utf-8")).hexdigest():
                if len(prefer_mfact_list) == 0:
                    top_prod_ids = self.popular_devices[:num_pick]
                else:
                    top_prod_ids = []
                    remains = []
                    for prod_id in self.popular_devices:
                        if self.products_meta[prod_id]["mfact_cd"] in prefer_mfact_list:
                            top_prod_ids.append(prod_id)
                        else:
                            remains.append(prod_id)
                    top_prod_ids += remains
                    top_prod_ids = top_prod_ids[:num_pick]
            else:
                user_meta = self.device_meta.get(user_eqp_mdl_cd, {})
                user_product_grp_id = user_meta.get("product_grp_id", "")

                if user_product_grp_id in self.target_prod_id:
                    y[self.target_prod_id.index(user_product_grp_id)] = 0

                if prefer_mfact_list:
                    for i, prod_id in enumerate(self.target_prod_id):
                        if self.products_meta[prod_id]["mfact_cd"] not in prefer_mfact_list:
                            y[i] = 0
                sort_labels = np.argsort(y)[::-1].tolist()
                top_prod_ids = [self.target_prod_id[label] for label in sort_labels[:num_pick]]

            items = []
            for prod_id in top_prod_ids:
                context_features = self.products.get(prod_id, {}).get("context_features", [])
                context_id = self.default_context_value
                if context_features:
                    for context_feature in context_features:
                        if re.match("^context_age[0-9]{2}", context_feature):
                            try:
                                age_bin = int(context_feature.split("_")[1][3:5])
                                if int(user_age) // 10 * 10 == age_bin:
                                    context_id = context_feature
                                    break
                            except Exception:
                                break
                        else:
                            context_id = context_feature
                            break

                context_meta = self.context_meta[self.context_meta["ContextId"] == context_id]
                text_id = context_meta["Text ID"].values[0] if len(context_meta) > 0 else ""
                text_value = context_meta["T world text"].values[0] if len(context_meta) > 0 else ""
                items.append(
                    {
                        "id": prod_id,
                        "name": self.products_meta[prod_id]["product_grp_nm"],
                        "props": {
                            "network_type": self.products_meta[prod_id]["eqp_mdl_ntwk"],
                            "product_id": self.products_meta[prod_id]["product_id"],
                            "color_hex": self.products_meta[prod_id]["color_hex"],
                            "color_nm": self.products_meta[prod_id]["color_nm"],
                            "text_id": text_id,
                            "text_value": text_value,
                        },
                        "type": "commbot_device",
                    }
                )

            return items
        except Exception as e:
            raise MLSModelError(f"CommbotCatboostDeviceModel: 후처리에 실패했습니다. {e}")

    def _convert(self, features):
        feature_names = self.features[3 : len(self.features) - len(self.default_emb_feature_values)]
        converted_features = []
        for i, f in enumerate(features):
            feature_name = feature_names[i]
            formula = self.conversion_formulas.get(feature_name)
            if formula:
                if "map" in formula:
                    converted_features.append(formula["map"].get(f, formula["default"]))
                elif "conditions" in formula:
                    conditions = formula["conditions"]
                    conditions_checked = True
                    for condition in conditions:
                        if (
                            (condition[0] == "==" and f != condition[1])
                            or (condition[0] == "!=" and f == condition[1])
                            or (condition[0] == "<" and f >= condition[1])
                            or (condition[0] == "<=" and f > condition[1])
                            or (condition[0] == ">" and f <= condition[1])
                            or (condition[0] == ">=" and f < condition[1])
                        ):
                            conditions_checked = False
                            break
                    converted_features.append(f if conditions_checked else formula["default"])
            else:
                try:
                    converted_feature = float({"Y": 1, "N": 0}.get(f, f))
                except Exception:
                    converted_feature = {"Y": 1, "N": 0}.get(f, f)

                converted_features.append(converted_feature)
        return converted_features



from .automl_helper import AutoMLHelper
from .automl_prediction import (
    AutoMLBatchPredictionClient,
    EmailChannelConfig,
    TOSChannelConfig,
    EDDChannelConfig,
    YEChannelConfig,
    ScoreExportOption,
    CustomFeatureSourceConfig,
    ExportChannelConfig,
    AutoMLBatchPrediction,
)

__all__ = [
    "AutoMLHelper",
    "AutoMLBatchPredictionClient",
    "EmailChannelConfig",
    "TOSChannelConfig",
    "EDDChannelConfig",
    "YEChannelConfig",
    "ScoreExportOption",
    "CustomFeatureSourceConfig",
    "ExportChannelConfig",
    "AutoMLBatchPrediction",
]



import pytest

import numpy as np

import lightgbm

from sktmls.models import MLSModelError
from sktmls.models.contrib import FeeNoEquipSingleEmbModel


class TestFeeNoEquipSingleEmbModels:
    @pytest.fixture(scope="class")
    def param(self):
        return {
            "model": FeeNoEquipSingleEmbModel(
                model=lightgbm.basic.Booster(train_set=lightgbm.Dataset(np.array([[1, 2]]), np.array([1]))),
                model_name="test_model",
                model_version="test_version",
                model_features=[
                    "age",
                    "bas_fee_amt",
                    "filter_five_g",
                    "additional_svc_ansim_option_scrb_type",
                    "embedding_vector",
                ],
                default_feature_values=[30, 100000, "Y", "free", [0.1, 0.2, 0.3, 0.4]],
                default_context_value="context_default",
                emb_feature_indices=[0, 2],
                conversion_formulas={
                    "additional_svc_ansim_option_scrb_type": {"map": {"free": 1, "paid": 2}, "default": 0},
                    "age": {"conditions": [(">", 0), ("<", 100)], "default": 30},
                },
                context_features=["filter_tfamilymoa", "context_data_overuse", "app_use_traffic_music_ratio_median_yn"],
                context_values=["tfamilymoa_value", "overuse_value", "music_ratio_value"],
                product_context_indices={
                    "NA00006404": [1, 2],
                    "NA00006536": [0],
                    "NA00006537": [0, 1],
                    "NA00006539": [0, 1, 2],
                },
            )
        }

    def test_000_no_rec_y(self, param, mocker):
        mocker.patch("lightgbm.basic.Booster.predict", return_value=np.array([0.0]))
        assert param.get("model").predict([30, 100000, "Y", "free", [0.1, 0.2, 0.3, 0.4], "Y", "Y", "Y"]) == {
            "items": []
        }

    def test_001_no_rec_context(self, param, mocker):
        mocker.patch("lightgbm.basic.Booster.predict", return_value=np.array([10.0]))
        assert param.get("model").predict([10, 100000, "N", "free", [0.1, 0.2, 0.3, 0.4], "Y", "Y", "Y"]) == {
            "items": []
        }
        assert param.get("model").predict([90, 100000, "N", "free", [0.1, 0.2, 0.3, 0.4], "Y", "Y", "Y"]) == {
            "items": []
        }

    def test_002_lte_upsell(self, param, mocker):
        mocker.patch("lightgbm.basic.Booster.predict", return_value=np.array([18.0]))
        assert param.get("model").predict([30, 30000, "N", "free", [0.1, 0.2, 0.3, 0.4], "Y", "Y", "Y"]) == {
            "items": [
                {
                    "id": "NA00006536",
                    "name": "T플랜 안심4G",
                    "type": "fee_no_equip",
                    "props": {"context_id": "tfamilymoa_value"},
                }
            ]
        }

        assert param.get("model").predict([30, 30000, "N", "free", [0.1, 0.2, 0.3, 0.4], "N", "Y", "Y"]) == {
            "items": [
                {
                    "id": "NA00006536",
                    "name": "T플랜 안심4G",
                    "type": "fee_no_equip",
                    "props": {"context_id": "context_default"},
                }
            ]
        }

    def test_003_5g_upsell(self, param, mocker):
        mocker.patch("lightgbm.basic.Booster.predict", return_value=np.array([18.0]))
        assert param.get("model").predict([30, 75000, "Y", "free", [0.1, 0.2, 0.3, 0.4], "Y", "N", "Y"]) == {
            "items": [
                {
                    "id": "NA00006404",
                    "name": "5GX 프라임",
                    "type": "fee_no_equip",
                    "props": {"context_id": "music_ratio_value"},
                }
            ]
        }
        assert param.get("model").predict([30, 75000, "Y", "free", [0.1, 0.2, 0.3, 0.4], "Y", "Y", "N"]) == {
            "items": [
                {
                    "id": "NA00006404",
                    "name": "5GX 프라임",
                    "type": "fee_no_equip",
                    "props": {"context_id": "overuse_value"},
                }
            ]
        }

    def test_004_none(self, param, mocker):
        mocker.patch("lightgbm.basic.Booster.predict", return_value=np.array([18.0]))
        assert param.get("model").predict([30, 75000, None, "free", [0.1, 0.2, 0.3, 0.4], "Y", "N", "N"]) == {
            "items": [
                {
                    "id": "NA00006404",
                    "name": "5GX 프라임",
                    "type": "fee_no_equip",
                    "props": {"context_id": "context_default"},
                }
            ]
        }

    def test_005_invalid_x(self, param, mocker):
        mocker.patch("lightgbm.basic.Booster.predict", return_value=np.array([18.0]))
        with pytest.raises(MLSModelError):
            param.get("model").predict([75000, "Y", "free", [0.1, 0.2, 0.3, 0.4]])
        with pytest.raises(MLSModelError):
            param.get("model").predict([30, 75000, "Y", "free", 0.1])
        with pytest.raises(MLSModelError):
            param.get("model").predict([30, 75000, "Y", "free", [0.1, 0.2, 0.3, 0.4, 0.5]])

    def test_006_conversion(self, param, mocker):
        mocker.patch("lightgbm.basic.Booster.predict", return_value=np.array([18.0]))
        assert param.get("model").predict([20000, 75000, "Y", "paid", [0.1, 0.2, 0.3, 0.4], "Y", "N", "N"]) == {
            "items": [
                {
                    "id": "NA00006404",
                    "name": "5GX 프라임",
                    "type": "fee_no_equip",
                    "props": {"context_id": "context_default"},
                }
            ]
        }



import random
from datetime import datetime
from typing import Any, Dict, List

from pytz import timezone

from sktmls.models import MLSRuleModel, MLSModelError

TZ = timezone("Asia/Seoul")
SEED_FORMAT = {"hourly": "%Y%m%d%H", "daily": "%Y%m%d", "weekly": "%Y%W", "monthly": "%Y%m", "yearly": "%Y"}


class MbrValidCategoryBaselineModel(MLSRuleModel):
    """
    MLS 모델 레지스트리에 등록되는 Rule 기반 클래스입니다.

    멤버십 단독 카테고리 유효카드 제공에 특화된 모델입니다.
    """

    def __init__(
        self,
        model_name: str,
        model_version: str,
        features: List[str],
        categories: List[str],
        item_type: str,
        num_pick: int = 1,
        random_cycle: str = "daily",
    ):
        """
        ## Args

        - model_name: (str) 모델 이름
        - model_version: (str) 모델 버전
        - features: (list(str)) 피쳐 리스트 (필수 피쳐: `user_id`)
        - categories: (list(str)) 전체 카테고리 리스트. `user_id` 이후 피쳐 순서와 동일해야 합니다.
        - item_type: (str) 결과의 `type`에 표기할 값
        - num_pick: (optional) (int) 랜덤 추출할 갯수 (기본값: 1)
        - random_cycle: (optional) (str) 랜덤 시드 초기화 주기 (`hourly`|`daily`|`weekly`|`monthly`|`yearly`) (기본값: `daily`)

        ## Example

        ```python
        my_model_v1 = MbrValidCategoryBaselineModel(
            model_name="my_model",
            model_version="v1",
            features=["user_id", "available_benefit_activity", "available_benefit_asianfood", "available_benefit_beauty"],
            categories=["mbr_activity", "mbr_asianfood", "mbr_beauty"],
            item_type="mbr_category_valid",
            num_pick=6,
            random_cycle="daily"
        )

        result = my_model_v1.predict(["1234567890", ["a1", "a2", "a3", "a4"], None, ["b1", "b2", "3", "4"]])
        ```
        """
        assert (
            isinstance(features, list) and len(features) > 0 and features[0] == "user_id"
        ), "`features`가 유효하지 않습니다. `features`의 첫 값은 반드시 `user_id`이어야 합니다."
        assert isinstance(categories, list) and len(categories) + 1 == len(
            features
        ), "전체 카테고리 리스트 길이가 필수 피쳐를 제외한 피쳐의 수와 다릅니다."
        assert isinstance(num_pick, int) and num_pick > 0, "`num_pick`이 유효하지 않습니다. 반드시 1 이상의 값을 가져야 합니다."
        assert random_cycle in SEED_FORMAT, "허용되지 않는 `random_cycle`값입니다."

        super().__init__(model_name, model_version, features)

        self.categories = categories
        self.item_type = item_type
        self.num_pick = num_pick
        self.random_cycle = random_cycle

    def predict(self, x: List[Any], **kwargs) -> Dict[str, Any]:
        for i, f in enumerate(x[1:]):
            if f is not None and not isinstance(f, list):
                raise MLSModelError(f"MbrValidCategoryBaselineModel: {self.features[i + 1]}이 리스트 형식으로 전달되지 않았습니다.")

        user_id = x[0]
        available_benefit = {
            category: x[i + 1]
            for i, category in enumerate(self.categories)
            if isinstance(x[i + 1], list) and len(x[i + 1]) > 1
        }.keys()

        random.seed(f"{user_id}{self.item_type}{datetime.now(TZ).strftime(SEED_FORMAT[self.random_cycle])}")

        if not available_benefit:
            return {"items": []}

        return {
            "items": [
                {
                    "id": self.item_type,
                    "name": "유효카테고리채널리스트",
                    "type": self.item_type,
                    "props": {
                        self.item_type: random.sample(available_benefit, min(self.num_pick, len(available_benefit)))
                    },
                }
            ]
        }



from .default_model import DefaultLightGBMModel, DefaultXGBoostModel, DefaultCatBoostModel, DefaultGenericModel
from .sample_model import SampleModel
from .sample_rule_model import SampleRuleModel
from .random_pick_model import RandomPickModel, PeriodicRandomPickModel
from .info_unpaid_rule_model import InfoUnpaidRuleModel
from .mbr_boostpark_baseline_model import MbrBoostparkBaselineModel
from .mbr_category_baseline_model import MbrCategoryBaselineModel
from .mbr_interest_all_baseline_model import MbrInterestAllBaselineModel
from .mbr_similar_benefit_model import MbrSimilarBenefitModel
from .mbr_popular_benefit_model import MbrPopularBenefitModel
from .mbr_custom_benefit_random_model import MbrCustomBenefitRandomModel
from .mbr_valid_category_baseline_model import MbrValidCategoryBaselineModel
from .info_defect_rule_model import InfoDefectRuleModel
from .shuffle_list_model import ShuffleListModel, PeriodicShuffleListModel
from .lightgbm_random_context_model import LightGBMRandomContextModel
from .vas_xcloud_rule_model import VasXcloudRuleModel
from .fee_no_equip_single_emb_model import FeeNoEquipSingleEmbModel
from .random_score_model import RandomScoreModel, PeriodicRandomScoreModel
from .generic_context_model import GenericContextModel
from .generic_logic_model import GenericLogicModel
from .conditional_generic_logic_model import ConditionalGenericLogicModel
from .tw_random_greeting_rule_model import TwRandomGreetingRuleModel
from .lightgbm_device_model import LightGBMDeviceModel
from .catboost_pointwise_ranking_model import CatBoostPointwiseRankingModel
from .mab_ranking_rule_model import MABRankingRuleModel
from .sample_pytorch_model import SamplePyTorchModel
from .catboost_device_model import CatboostDeviceModel
from .commbot_catboost_device_model import CommbotCatboostDeviceModel

__all__ = [
    "DefaultLightGBMModel",
    "DefaultXGBoostModel",
    "DefaultCatBoostModel",
    "DefaultGenericModel",
    "SampleModel",
    "SampleRuleModel",
    "RandomPickModel",
    "PeriodicRandomPickModel",
    "InfoUnpaidRuleModel",
    "MbrBoostparkBaselineModel",
    "MbrCategoryBaselineModel",
    "MbrInterestAllBaselineModel",
    "MbrSimilarBenefitModel",
    "MbrPopularBenefitModel",
    "MbrCustomBenefitRandomModel",
    "MbrValidCategoryBaselineModel",
    "InfoDefectRuleModel",
    "ShuffleListModel",
    "PeriodicShuffleListModel",
    "LightGBMRandomContextModel",
    "VasXcloudRuleModel",
    "FeeNoEquipSingleEmbModel",
    "RandomScoreModel",
    "PeriodicRandomScoreModel",
    "GenericContextModel",
    "GenericLogicModel",
    "ConditionalGenericLogicModel",
    "TwRandomGreetingRuleModel",
    "LightGBMDeviceModel",
    "CatBoostPointwiseRankingModel",
    "MABRankingRuleModel",
    "SamplePyTorchModel",
    "CatboostDeviceModel",
    "CommbotCatboostDeviceModel",
]



from typing import Dict, Any, List

from sktmls import MLSClient, MLSENV, MLSRuntimeENV, MLSClientError, MLSResponse

MLS_VERTEX_API_URL = "/api/v1/vertices"
MLS_EDGE_API_URL = "/api/v1/edges"


class VertexSchema:
    """
    MLS Vertex 스키마 클래스입니다.
    """

    def __init__(self, **kwargs):
        """
        ## Args

        - kwargs
            - id: (str) Vertex 스키마 ID
            - label: (str) Vertex 스키마 라벨
            - properties: (dict) Vertex property 데이터 타입 정의
        ## Returns
        `sktmls.graphs.VertexSchema`
        """
        self.id = kwargs.get("id")
        self.label = kwargs.get("label")
        self.properties = kwargs.get("properties")

    def get(self):
        return self.__dict__


class EdgeSchema:
    """
    MLS Edge 스키마 클래스입니다.
    """

    def __init__(self, **kwargs):
        """
        ## Args

        - kwargs
            - id: (str) Edge 스키마 ID
            - label: (str) Edge 스키마 라벨
            - frm: (str) 출발 Vertex 라벨
            - to: (str) 도착 Vertex 라벨
            - properties: (dict) Edge property 데이터 타입 정의

        ## Returns
        `sktmls.graphs.EdgeSchema`
        """
        self.id = kwargs.get("id")
        self.label = kwargs.get("label")
        self.frm = kwargs.get("frm")
        self.to = kwargs.get("to")
        self.properties = kwargs.get("properties")

    def get(self):
        return self.__dict__


class VertexClient(MLSClient):
    """
    MLS Vertex 스키마 관련 기능들을 제공하는 클라이언트입니다.
    """

    def __init__(
        self,
        env: MLSENV = None,
        runtime_env: MLSRuntimeENV = None,
        username: str = None,
        password: str = None,
    ):
        """
        ## Args

        - env: (`sktmls.MLSENV`) 접근할 MLS 환경 (`sktmls.MLSENV.DEV`|`sktmls.MLSENV.STG`|`sktmls.MLSENV.PRD`) (기본값: `sktmls.MLSENV.STG`)
        - runtime_env: (`sktmls.MLSRuntimeENV`) 클라이언트가 실행되는 환경 (`sktmls.MLSRuntimeENV.YE`|`sktmls.MLSRuntimeENV.EDD`|`sktmls.MLSRuntimeENV.LOCAL`) (기본값: `sktmls.MLSRuntimeENV.LOCAL`)
        - username: (str) MLS 계정명 (기본값: $MLS_USERNAME)
        - password: (str) MLS 계정 비밀번호 (기본값: $MLS_PASSWORD)

        ## Returns
        `sktmls.graphs.VertexClient`

        ## Example

        ```python
        vertex_client = VertexClient(env=MLSENV.STG, runtime_env=MLSRuntimeENV.YE, username="mls_account", password="mls_password")
        ```
        """
        super().__init__(env, runtime_env, username, password)

    def create_vertex(self, label: str, properties: Dict[str, str] = {}) -> VertexSchema:
        """
        Vertex 스키마를 생성합니다.

        ## Args

        - label: (str) Vertex 스키마 라벨
        - properties: (optional) (dict) Vertex 스키마의 데이터 속성 (`Float(single)`|`Float(set)`|`String(single)`|`String(set)`)
            - Float, String: 데이터 타입
            - single: 단일 property 값을 계속 업데이트
            - set: property값을 여러 개 계속 누적해 저장

        ## Returns
        `sktmls.graphs.VertexSchema`

        ## Example

        ```python
        vertex = vertex_client.create_vertex(
            label="SampleVertex",
            properties={
                "name":"String(single)",
                "lat":"Float(single)",
                "long": "Float(single)"
            }
        )
        ```
        """

        for p in properties.values():
            assert p in ["Float(single)", "Float(set)", "String(single)", "String(set)"], "허용되지 않은 property 타입입니다."
        data = {"label": label}
        if properties:
            data["properties"] = properties
        return VertexSchema(**self._request(method="POST", url=f"{MLS_VERTEX_API_URL}", data=data).results)

    def list_vertices(self, **kwargs) -> List[VertexSchema]:
        """
        vertex 스키마 리스트를 가져옵니다.

        ## Args

        - kwargs: (optional) (dict) 쿼리 조건
            - id: (int) vertex 스키마 id
            - label: (str) vertex 스키마 라벨
            - query: (str) 검색 문자
            - page: (int) 페이지 번호

        ## Returns
        list(`sktmls.graphs.VertexSchema`)

        ## Example

        ```python
        vertices = vertex_client.list_vertices()
        ```
        """
        response = self._request(method="GET", url=f"{MLS_VERTEX_API_URL}", params=kwargs).results
        return [VertexSchema(**vertex) for vertex in response]

    def get_vertex(self, id: int = None, label: str = None) -> VertexSchema:
        """
        단일 Vertex 스키마 정보를 가져옵니다.

        ## Args: `id` 또는 `label` 중 한 개 이상의 값이 반드시 전달되어야 합니다.

        - id: (int) vertex 스키마 id
        - label: (str) vertex 스키마 라벨

        ## Returns
        `sktmls.graphs.VertexSchema`

        ## Example

        ```python
        vertex = vertex_client.get_vertex(id=1)
        vertex = vertex_client.get_vertex(label="SampleVertex")
        ```
        """
        assert id or label, "`id` 또는 `label` 중 한 개 이상의 값이 반드시 전달되어야 합니다."

        vertices = self.list_vertices(id=id, label=label)
        if len(vertices) == 0:
            raise MLSClientError(code=404, msg="Vertex 스키마가 없습니다.")
        if len(vertices) > 1:
            raise MLSClientError(code=409, msg="Vertex 스키마가 여러개 존재합니다.")
        return vertices[0]

    def update_vertex(self, vertex: VertexSchema, label: str, properties: Dict[str, Any] = {}) -> VertexSchema:
        """
        단일 Vertex 스키마를 수정합니다.
        ## Args

        - vertex: (VertexSchema) Vertex 스키마 객체
        - label: (str) Vertex 스키마 라벨
        - properties: (optional) (dict) Vertex 스키마의 데이터 속성 (`Float(single)`|`Float(set)`|`String(single)`|`String(set)`)
            - Float, String: 데이터 타입
            - single: 단일 property 값을 계속 업데이트
            - set: property값을 여러 개 계속 누적해 저장

        ## Returns
        `sktmls.graphs.VertexSchema`

        ## Example

        ```python
        new_vertex = vertex_client.update_vertex(
            vertex,
            label="SampleVertex2",
            properties={
                "name":"String(single)"
            }
        )
        ```
        """
        for p in properties.values():
            assert p in ["Float(single)", "Float(set)", "String(single)", "String(set)"], "허용되지 않은 property 타입입니다."
        data = {"label": label}
        if properties:
            data["properties"] = properties
        return VertexSchema(**self._request(method="PUT", url=f"{MLS_VERTEX_API_URL}/{vertex.id}", data=data).results)

    def delete_vertex(self, vertex: VertexSchema) -> MLSResponse:
        """
        단일 Vertex 스키마를 삭제합니다. 연결된 Edge 스키마도 함께 삭제됩니다.

        ## Args

        - vertex: (VertexSchema) Vertex 스키마 객체

        ## Returns
        `sktmls.MLSResponse`

        ## Example

        ```python
        vertex_client.delete_vertex(vertex)
        ```
        """
        assert type(vertex) == VertexSchema
        return self._request(method="DELETE", url=f"{MLS_VERTEX_API_URL}/{vertex.id}")


class EdgeClient(MLSClient):
    """
    MLS Edge 스키마 관련 기능들을 제공하는 클라이언트입니다.
    """

    def __init__(
        self,
        env: MLSENV = None,
        runtime_env: MLSRuntimeENV.LOCAL = None,
        username: str = None,
        password: str = None,
    ):
        """
        ## Args

        - env: (`sktmls.MLSENV`) 접근할 MLS 환경 (`sktmls.MLSENV.DEV`|`sktmls.MLSENV.STG`|`sktmls.MLSENV.PRD`) (기본값: `sktmls.MLSENV.STG`)
        - runtime_env: (`sktmls.MLSRuntimeENV`) 클라이언트가 실행되는 환경 (`sktmls.MLSRuntimeENV.YE`|`sktmls.MLSRuntimeENV.EDD`|`sktmls.MLSRuntimeENV.LOCAL`) (기본값: `sktmls.MLSRuntimeENV.LOCAL`)
        - username: (str) MLS 계정명 (기본값: $MLS_USERNAME)
        - password: (str) MLS 계정 비밀번호 (기본값: $MLS_PASSWORD)

        ## Returns
        `sktmls.graphs.EdgeClient`

        ## Example

        ```python
        edge_client = EdgeClient(env=MLSENV.STG, runtime_env=MLSRuntimeENV.YE, username="mls_account", password="mls_password")
        ```
        """
        super().__init__(env, runtime_env, username, password)

    def create_edge(self, label: str, frm: str, to: str, properties: Dict[str, str] = {}) -> EdgeSchema:
        """
        Edge 스키마를 생성합니다.

        ## Args
        - label: (str) Edge 스키마 라벨
        - frm: (str) Edge 시작 Vertex 라벨
        - to: (str) Edge 종료 Vertex 라벨
        - properties: (optional) (dict) Vertex 스키마의 데이터 속성 (`Float(single)`|`Float(set)`|`String(single)`|`String(set)`)
            - Float, String: 데이터 타입
            - single: 단일 property 값을 계속 업데이트
            - set: property값을 여러 개 계속 누적해 저장

        ## Returns
        `sktmls.graphs.EdgeSchema`

        ## Example

        ```python
        edge = edge_client.create_edge(
            label="SampleEdge",
            frm="startVertex",
            to="endVertex",
            properties={
                "name":"String(single)",
                "lat":"Float(single)",
                "long": "Float(single)"
            }
        )
        ```
        """

        for p in properties.values():
            assert p in ["Float(single)", "Float(set)", "String(single)", "String(set)"], "허용되지 않은 property 타입입니다."
        data = {"label": label, "frm": frm, "to": to}
        if properties:
            data["properties"] = properties
        return EdgeSchema(**self._request(method="POST", url=f"{MLS_EDGE_API_URL}", data=data).results)

    def list_edges(self, **kwargs) -> List[EdgeSchema]:
        """
        Edge 스키마 리스트를 가져옵니다.

        ## Args

        - kwargs: (optional) (dict) 쿼리 조건
            - id: (int) Edge 스키마 id
            - label: (str) Edge 스키마 라벨
            - query: (str) 검색 문자
            - page: (int) 페이지 번호

        ## Returns
        list(`sktmls.graphs.EdgeSchema`)

        ## Example

        ```python
        edges = edge_client.list_edges()
        ```
        """
        response = self._request(method="GET", url=f"{MLS_EDGE_API_URL}", params=kwargs).results
        return [EdgeSchema(**edge) for edge in response]

    def get_edge(self, id: int = None, label: str = None) -> EdgeSchema:
        """
        단일 Edge 스키마 정보를 가져옵니다.

        ## Args: `id` 또는 `label` 중 한 개 이상의 값이 반드시 전달되어야 합니다.

        - id: (int) Edge 스키마 id
        - label: (str) Edge 스키마 라벨

        ## Returns
        `sktmls.graphs.EdgeSchema`

        ## Example

        ```python
        edge = edge_client.get_edge(id=1)
        edge = edge_client.get_edge(label="SampleEdge")
        ```
        """
        assert id or label, "`id` 또는 `label` 중 한 개 이상의 값이 반드시 전달되어야 합니다."

        edges = self.list_edges(id=id, label=label)
        if len(edges) == 0:
            raise MLSClientError(code=404, msg="Edge 스키마가 없습니다.")
        if len(edges) > 1:
            raise MLSClientError(code=409, msg="Edge 스키마가 여러개 존재합니다.")
        return edges[0]

    def update_edge(
        self, edge: EdgeSchema, label: str, frm: str, to: str, properties: Dict[str, Any] = {}
    ) -> EdgeSchema:
        """
        단일 Edge 스키마를 수정합니다.

        ## Args

        - edge: (EdgeSchema) Edge 스키마 객체
        - label: (str) Edge 스키마 라벨
        - frm: (str) Edge 시작 Vertex 라벨
        - to: (str) Edge 종료 Vertex 라벨
        - properties: (optional) (dict) Vertex 스키마의 데이터 속성 (`Float(single)`|`Float(set)`|`String(single)`|`String(set)`)
            - Float, String: 데이터 타입
            - single: 단일 property 값을 계속 업데이트
            - set: property값을 여러 개 계속 누적해 저장

        ## Returns
        `sktmls.graphs.EdgeSchema`

        ## Example

        ```python
        edge_client.update_edge(
            edge,
            label="SampleEdge2",
            frm="startVertex",
            to="endVertex",
            properties={
                "name":"String(single)"
            }
        )
        ```
        """
        for p in properties.values():
            assert p in ["Float(single)", "Float(set)", "String(single)", "String(set)"], "허용되지 않은 property 타입입니다."
        data = {"label": label, "frm": frm, "to": to}
        if properties:
            data["properties"] = properties
        return EdgeSchema(**self._request(method="PUT", url=f"{MLS_EDGE_API_URL}/{edge.id}", data=data).results)

    def delete_edge(self, edge: EdgeSchema) -> MLSResponse:
        """
        단일 Edge 스키마를 삭제합니다.

        ## Args

        - edge: (EdgeSchema) Edge 스키마 객체

        ## Returns
        `sktmls.MLSResponse`

        ## Example

        ```python
        edge_client.delete_edge(edge)
        ```
        """
        assert type(edge) == EdgeSchema
        return self._request(method="DELETE", url=f"{MLS_EDGE_API_URL}/{edge.id}")



import aiohttp
import base64
import json
import os
import requests
from typing import Any, Dict

from sktmls import MLSENV, MLSRuntimeENV
from sktmls.config import Config


class MLSClientError(Exception):
    def __init__(self, code: int, msg: str):
        super().__init__(msg)
        self.code = code
        self.msg = msg


class DoesNotExist(MLSClientError):
    def __init__(self, msg="결과가 존재하지 않습니다"):
        super().__init__(404, msg)


class MLSResponse:
    def __init__(self, response: requests.models.Response = None, status: int = None, body: Dict[str, Any] = None):
        if response is not None and status is None and body is None:
            data = json.loads(response.content.decode("utf-8"))
            self.status = response.status_code
        else:
            data = body
            self.status = status
        self.code = data.get("code")
        self.error = data.get("error")
        self.results = data.get("results")

        if self.error:
            raise MLSClientError(code=self.code, msg=self.error)


class MLSClient:
    def __init__(
        self,
        env: MLSENV = None,
        runtime_env: MLSRuntimeENV.LOCAL = None,
        username: str = None,
        password: str = None,
    ):
        if env:
            assert env in MLSENV.list_items(), "유효하지 않은 MLS 환경입니다."
            self.__env = env
        elif os.environ.get("MLS_ENV"):
            self.__env = MLSENV[os.environ["MLS_ENV"]]
        elif os.environ.get("AWS_ENV"):
            self.__env = MLSENV[os.environ["AWS_ENV"]]
        else:
            self.__env = MLSENV.STG

        if runtime_env:
            assert runtime_env in MLSRuntimeENV.list_items(), "유효하지 않은 런타임 환경입니다."
            self.__runtime_env = runtime_env
        elif os.environ.get("MLS_RUNTINE_ENV"):
            self.__runtime_env = MLSRuntimeENV[os.environ["MLS_RUNTINE_ENV"]]
        else:
            __HOSTNAME = os.environ.get("HOSTNAME", "").lower()
            if __HOSTNAME.startswith("bdp-dmi"):
                self.__runtime_env = MLSRuntimeENV.YE
            elif __HOSTNAME.startswith("vm-skt"):
                self.__runtime_env = MLSRuntimeENV.EDD
            else:
                self.__runtime_env = MLSRuntimeENV.LOCAL

        self.config = Config(self.__runtime_env.value)

        if runtime_env != MLSRuntimeENV.MMS:
            if username:
                assert type(username) == str, "`username`의 타입이 str이 아닙니다."
                self.__username = username
            elif os.environ.get("MLS_USERNAME"):
                self.__username = os.environ["MLS_USERNAME"]
            else:
                raise MLSClientError(code=400, msg="`username`이 전달되지 않았습니다. 파라미터 또는 $MLS_USERNAME 환경 변수를 확인해주세요.")

            if password:
                assert type(password) == str, "`password`의 타입이 str이 아닙니다."
                self.__password = password
            elif os.environ.get("MLS_PASSWORD"):
                self.__password = os.environ["MLS_PASSWORD"]
            else:
                raise MLSClientError(code=400, msg="`password`가 전달되지 않았습니다. 파라미터 또는 $MLS_PASSWORD 환경 변수를 확인해주세요.")

            self.__api_token = base64.b64encode(f"{self.__username}:{self.__password}".encode()).decode("utf-8")
        else:
            self.__api_token = "None"

    def get_env(self) -> MLSENV:
        return self.__env

    def get_runtime_env(self) -> MLSRuntimeENV:
        return self.__runtime_env

    def get_username(self) -> str:
        return self.__username

    def _request(self, method: str, url: str, data: Any = None, params=None) -> MLSResponse:
        response = requests.request(
            method=method,
            url=f"{self.config.MLS_AB_API_URL[self.__env.value]}{'' if url.startswith('/') else '/'}{url}",
            headers={"Authorization": f"Basic {self.__api_token}"},
            json=data,
            params=params,
        )

        return MLSResponse(response)

    async def _request_async(self, method: str, url: str, data: Any = None, params=None) -> MLSResponse:
        async with aiohttp.ClientSession() as session:
            async with session.request(
                method=method,
                url=f"{self.config.MLS_AB_API_URL[self.__env.value]}{'' if url.startswith('/') else '/'}{url}",
                headers={"Authorization": f"Basic {self.__api_token}"},
                json=data,
                params=params,
            ) as response:
                return MLSResponse(status=response.status, body=await response.json())
